import cv2
import numpy as np
import torch
from torchvision import transforms
from torch2trt import TRTModule
import os


def imgs2tensor(imgs):
    """
    imgs: list of images, each image has shape (28,28)
    """
    mean = (0.1307, )
    std = (0.3081, )
    tensors = []

    for img in imgs:
        img = np.expand_dims(img, axis=-1)
        img = transforms.functional.to_tensor(img)
        img = transforms.functional.normalize(img, mean, std)
        tensors.append(img)
    return torch.stack(tensors)


def run_inference(model, imgs):
    """
    imgs: list of images, each image has shape (28,28)
    """
    imgs = imgs2tensor(imgs).cuda().half()
    with torch.no_grad():
        output = model(imgs)
        scores, predict_num = torch.max(output, 1)
    return scores, predict_num


def threshold_img(frame):
    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)  # 将图像转为灰度图像
    gray = cv2.GaussianBlur(gray, (5, 5), 0)  # 应用高斯模糊去噪声
    _, th = cv2.threshold(gray, 90, 255,
                          cv2.THRESH_BINARY_INV)  # 将亮度低于90的像素值设为255，其它设为0
    return th


def find_contours(th_im):
    cnts, _ = cv2.findContours(th_im, cv2.RETR_EXTERNAL,
                               cv2.CHAIN_APPROX_SIMPLE)
    # 将面积小于133的轮廓忽略掉，该数值可酌情修改
    cnts = [cnt for cnt in cnts if cv2.contourArea(cnt) > 133]
    return cnts


def white_black_ratio(roi):
    """
    计算白色像素与黑色像素之比
    """
    hist = cv2.calcHist([roi], [0], None, [2], [0, 256])
    extent = float(hist[1]) / hist[0]
    return extent


def get_digit_rect(cnts, th_im):
    rects = [cv2.boundingRect(cnt) for cnt in cnts]
    rects = [
        rect for rect in rects
        if white_black_ratio(th_im[rect[1]:rect[1] + rect[3],
                                   rect[0]:rect[0] + rect[2]]) < 1
    ]
    return rects


def get_digit_imgs(th, rects):
    roi = [
        th[rect[1]:rect[1] + rect[3], rect[0]:rect[0] + rect[2]]
        for rect in rects
    ]
    resize = []
    for i in roi:
        h, w = i.shape
        # 缩放矩形区域，使其最大边长为18个像素
        scale = 18 / h if h > w else 18 / w
        i = cv2.resize(i, (0, 0), fx=scale, fy=scale)
        resize.append(i)

    # 将含有数字像素的图像复制到28x28的黑色背景中
    mnist_digits = []
    for d in resize:
        black = np.zeros((28, 28), dtype=np.uint8)
        h, w = d.shape
        black[int(14 - h / 2):int(14 + h / 2),
              int(14 - w / 2):int(14 + w / 2)] = d
        mnist_digits.append(black)
    return mnist_digits


def draw_predict_num(frame, scores, predict_num, rects):
    font = cv2.FONT_HERSHEY_SIMPLEX
    for score, num, rect in zip(scores, predict_num, rects):
        # 成绩小于0.8的当无效
        if score > 0.8:
            cv2.putText(frame, '%d: %.1f%%' % (num, score * 100),
                        (rect[0], rect[1]), font, 1, (255, 255, 0), 3,
                        cv2.LINE_AA)
            cv2.rectangle(frame, (rect[0], rect[1]),
                          (rect[0] + rect[2], rect[1] + rect[3]), (0, 255, 0),
                          3)


device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

cur_dir = os.path.split(os.path.realpath(__file__))[0]


class DigitDetector():
    def __init__(self) -> None:
        self.model = TRTModule()
        self.model.load_state_dict(
            torch.load(os.path.join(cur_dir, 'mnist_trt_b9.pth')))
        self.model = self.model.to(device)
        self.model.eval()
        x = torch.ones(1, 1, 28, 28).cuda().half()
        self.model(x)

    def detect(self, frame):
        th = threshold_img(frame)
        height, width = th.shape
        cnts = find_contours(th)
        rects = get_digit_rect(cnts, th)
        rects = [
            r for r in rects if r[2] > 40 and r[3] > 40 and r[2] < height /
            2 and r[3] < height / 2
        ]
        digits = get_digit_imgs(th, rects)
        if len(digits):
            # 最多送入 9 张图片，因为 mnist_trt_b9.pth 设置了最大 batchSize 为 9
            scores, predict_num = run_inference(self.model, digits[:9])
            # 将结果显示在frame中
            draw_predict_num(frame, scores, predict_num, rects)



import cv2
import torch
import PIL.Image
from torch2trt import TRTModule
from fro_AI.utils.data_utils import get_file, get_hash_prefix_from_file_name
import torch.nn.functional as F
import torchvision.transforms as transforms

model_url = 'http://gz.chuangfeigu.com:8087/fro_AI/models/ObstacleClassifier/obstacle_trt_70fbe733.pth'

device = torch.device('cuda')

mean = torch.Tensor([0.485, 0.456, 0.406]).cuda().half()
std = torch.Tensor([0.229, 0.224, 0.225]).cuda().half()


def preprocess(image):
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    image = PIL.Image.fromarray(image)
    image = transforms.functional.to_tensor(image).to(device).half()
    image = transforms.functional.resize(image, [224, 224])
    image.sub_(mean[:, None, None]).div_(std[:, None, None])
    return image[None, ...]  # 增加一个维度


class ObstacleClassifier():
    def __init__(self) -> None:
        self.model = TRTModule()
        fname = 'obstacle_trt_70fbe733.pth'
        hash_prefix = get_hash_prefix_from_file_name(fname)
        model_path = get_file(fname,
                              model_url,
                              hash_prefix=hash_prefix,
                              cache_subdir='models/ObstacleClassifier')
        self.model.load_state_dict(torch.load(model_path))
        x = torch.ones(1, 1, 224, 224).cuda().half()
        self.model(x)

    def predict(self, frame):
        """predict block or free

        Parameters
        ----------
        frame : np.ndarray
            bgr picture

        Returns
        -------
        float
            1: blocked; 0: free
        """
        x = preprocess(frame)
        y = self.model(x)
        y = F.softmax(y, dim=1)
        return float(y.flatten()[0])



import cv2
import jetson.inference
import numpy as np
from numpy import linalg as LA
import os


class ObjectDetector():
    def __init__(self) -> None:
        self.net = jetson.inference.detectNet('ssd-mobilenet-v2')
        x = np.ones((224, 224, 3))
        cuda_mem = jetson.utils.cudaFromNumpy(x)
        self.net.Detect(cuda_mem)
        del cuda_mem
        self.target = 'None'
        cur_dir = os.path.split(os.path.realpath(__file__))[0]
        path_to_labels = os.path.join(cur_dir, 'ssd_coco_labels.txt')
        label_list = []
        self.label_map_dict = {}
        with open(path_to_labels, 'r') as f:
            label_list = [line[:-1] for line in f]
            self.label_map_dict = {v: i for i, v in enumerate(label_list)}
        self.label_map_dict['None'] = 999
        self.height = 360
        self.width = 640

    def detect(self, frame):
        self.height, self.width = frame.shape[:2]
        rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        cuda_mem = jetson.utils.cudaFromNumpy(rgb)
        detections = self.net.Detect(cuda_mem)
        res = jetson.utils.cudaAllocMapped(width=cuda_mem.width,
                                           height=cuda_mem.height,
                                           format='bgr8')
        jetson.utils.cudaConvertColor(cuda_mem, res)
        jetson.utils.cudaDeviceSynchronize()
        frame[:] = jetson.utils.cudaToNumpy(res)[:]
        return detections

    def set_target(self, target):
        if target not in self.label_map_dict.keys():
            self.target = 'None'
        else:
            self.target = target

    def get_target_location(self, detections):
        centers = []
        for detection in detections:
            if self.label_map_dict[
                    self.
                    target] == detection.ClassID and detection.Confidence > 0.5:
                centers.append(detection.Center)
        return self.closest_detection(centers)

    def closest_detection(self, centers):
        closest_detection = (0, 0)
        ret = False
        min_distance = 100
        for c in centers:
            cn = [c[0] / self.width * 2 - 1, c[1] / self.height * 2 - 1]
            distance = LA.norm(cn)
            if not ret:
                closest_detection = cn
                min_distance = distance
                ret = True
            elif distance < min_distance:
                closest_detection = cn
                min_distance = distance
        return ret, closest_detection[0], closest_detection[1]



import cv2.aruco as aruco
import numpy as np

default_cameraMatrix = np.array([[392.75533375, 0., 284.73851876],
                                 [0., 413.9101659, 254.38799118], [0., 0.,
                                                                   1.]])
default_distCoeffs = np.array(
    [[0.0290911, -1.24869976, -0.01714862, -0.02837211, 3.59261617]])

aruco_dict = aruco.custom_dictionary(10, 6)


class ArUcoDetector():
    def __init__(self, marker_size=0.06) -> None:
        self.marker_size = marker_size

    def detect(self, bgr):
        ret = False
        x, y, z = 0, 0, 0
        corners, ids, _ = aruco.detectMarkers(bgr, aruco_dict)
        if ids is not None:
            _, tvecs, _ = aruco.estimatePoseSingleMarkers(
                corners, self.marker_size, default_cameraMatrix,
                default_distCoeffs)
            x = tvecs[0][0][0].item()
            y = tvecs[0][0][1].item()
            z = tvecs[0][0][2].item()
            if z > 0.05:
                aruco.drawDetectedMarkers(bgr, corners, ids)
                ret = True
        return ret, x, y, z



import cv2
import numpy as np
from simple_pid import PID


def get_color_mask(bgr, lower, upper):
    """获取颜色遮罩

    Parameters
    ----------
    bgr : numpy.ndarray
        bgr格式的图像
    lower : numpy.ndarray
        颜色下界，HSV表示
    upper : numpy.ndarray
        颜色上界，HSV表示

    Returns
    -------
    numpy.ndarray
        颜色遮罩
    """

    # 利用高斯模糊减少图像噪声
    res = cv2.GaussianBlur(bgr, (5, 5), 0)

    # 将图像转为HSV格式，因为HSV可以更方便表示颜色
    res = cv2.cvtColor(res, cv2.COLOR_BGR2HSV)

    # 利用`cv2.inRange()`函数创建颜色遮罩
    res = cv2.inRange(res, lower, upper)

    return res


class ColorTracker():
    def __init__(self, lower_hsv, upper_hsv, min_r) -> None:
        self.lower_hsv = np.array(lower_hsv)
        self.upper_hsv = np.array(upper_hsv)
        self.min_r = min_r

    def find_max_color_blob(self, bgr):
        """获取颜色坐标

        Parameters
        ----------
        bgr : numpy.ndarray
            bgr 格式的图像

        Returns
        -------
        x : int
        y : int
        r : int
        """
        mask = get_color_mask(bgr, self.lower_hsv, self.upper_hsv)
        cnts, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL,
                                   cv2.CHAIN_APPROX_SIMPLE)
        x, y, r = -1, -1, -1
        ret = False
        if len(cnts) > 0:
            # 查找最大的轮廓
            c = max(cnts, key=cv2.contourArea)

            # 利用该轮廓计算其外接圆形
            (x, y), r = cv2.minEnclosingCircle(c)
            x, y, r = int(x), int(y), int(r)
            if r > self.min_r:
                cv2.circle(bgr, (x, y), r, (255, 0, 0), 2)  # 在BGR图像上绘制圆
                ret = True
        h, w = bgr.shape[:2]
        x /= w
        y /= h
        return ret, x, y, r

    def set_x_pid(self,
                  kp,
                  ki=0,
                  kd=0,
                  setpoint=0,
                  output_limits=(None, None)):
        self.x_pid = PID(kp,
                         ki,
                         kd,
                         setpoint=setpoint,
                         output_limits=output_limits)

    def set_y_pid(self,
                  kp,
                  ki=0,
                  kd=0,
                  setpoint=0,
                  output_limits=(None, None)):
        self.y_pid = PID(kp,
                         ki,
                         kd,
                         setpoint=setpoint,
                         output_limits=output_limits)

    def set_z_pid(self,
                  kp,
                  ki=0,
                  kd=0,
                  setpoint=0,
                  output_limits=(None, None)):
        self.z_pid = PID(kp,
                         ki,
                         kd,
                         setpoint=setpoint,
                         output_limits=output_limits)

    def get_x_output(self, detection):
        if not hasattr(self, 'x_pid') or not detection[0]:
            return 0

        error = detection[1] - 0.5
        return self.x_pid(error)

    def get_y_output(self, detection):
        if not hasattr(self, 'y_pid') or not detection[0]:
            return 0
        error = detection[2] - 0.5
        return self.y_pid(error)

    def get_z_output(self, detection):
        if not hasattr(self, 'z_pid') or not detection[0]:
            return 0
        return self.z_pid(detection[3])



import cv2
import torch
import PIL.Image
from torch2trt import TRTModule
from fro_AI.utils.data_utils import get_file, get_hash_prefix_from_file_name
import torch.nn.functional as F
import torchvision.transforms as transforms
import numpy as np

model_url = 'http://gz.chuangfeigu.com:8087/fro_AI/models/road/road_trt_5aac4006.pth'

device = torch.device('cuda')

mean = torch.Tensor([0.485, 0.456, 0.406]).cuda().half()
std = torch.Tensor([0.229, 0.224, 0.225]).cuda().half()


def preprocess(image):
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    image = PIL.Image.fromarray(image)
    image = transforms.functional.to_tensor(image).to(device).half()
    image = transforms.functional.resize(image, [224, 224])
    image.sub_(mean[:, None, None]).div_(std[:, None, None])
    return image[None, ...]  # 增加一个维度


class RoadFollower():
    def __init__(self) -> None:
        self.model = TRTModule()
        fname = 'road_trt_5aac4006.pth'
        hash_prefix = get_hash_prefix_from_file_name(fname)
        model_path = get_file(fname,
                              model_url,
                              hash_prefix=hash_prefix,
                              cache_subdir='models/road')
        self.model.load_state_dict(torch.load(model_path))
        x = torch.ones(1, 1, 224, 224).cuda().half()
        self.model(x)

    def predict(self, frame):
        xy = self.model(
            preprocess(frame)).detach().float().cpu().numpy().flatten()
        x = xy[0]
        y = xy[1]
        angle = np.arctan2(x, y)
        h, w = frame.shape[:2]
        cx = int(w * (x / 2 + 0.5))
        cy = int(h * y)
        cv2.circle(frame, (cx, cy), 10, (0, 255, 0), 2, cv2.LINE_AA)
        return angle



def clamp(value, lower=None, upper=None):
    if value is None:
        return None
    elif (upper is not None) and (value > upper):
        return upper
    elif (lower is not None) and (value < lower):
        return lower
    return value



import setuptools
import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="jetbot_scratch",  # Replace with your own username
    version=get_version("jetbot_scratch/__init__.py"),
    author="z14git",
    author_email="lzl1992@gmail.com",
    description="Jetbot 积木编程包",
    url="https://gitee.com/z14git/jetbot_scratch",
    extras_require={
        'all': [],
    },
    install_requires=['simple-pid'],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    package_data={
        'jetbot_scratch': ['*.pth', '*.txt']
    }
)
